/**HEADER********************************************************************
* 
* Copyright (c) 2008 Freescale Semiconductor;
* All Rights Reserved
*
* Copyright (c) 1989-2008 ARC International;
* All Rights Reserved
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: host_rcv.c$
* $Version : 3.0.3.0$
* $Date    : Sep-22-2008$
*
* Comments:
*
*   This file contains the USB Host API specific functions to receive data.
*
*END************************************************************************/
#include "hostapi.h"
#include "usbprv_host.h"
#include "host_rcv.h"

#ifdef __USB_OS_MQX__
   #include "mqx_arc.h"
#endif

/*FUNCTION*-------------------------------------------------------------
*
*  Function Name  : _usb_host_recv_data
*  Returned Value : error or status of the transfer
*  Comments       :
* The Receive Data routine is non-blocking routine that causes a buffer 
* to be made available for data recieved from the 
* USB host. It takes the buffer and passes it down to lower
level hardware driver. 
*END*-----------------------------------------------------------------*/
USB_STATUS _usb_host_recv_data
   (
      /* [IN] the USB Host state structure */
      _usb_host_handle           handle, 

      /* [IN] the pipe handle */
      _usb_pipe_handle           pipe_handle,

      /* [IN] transfer parameters */
      TR_INIT_PARAM_STRUCT_PTR   tr_params_ptr
   )
{ /* Body */
   PIPE_DESCRIPTOR_STRUCT_PTR pipe_descr_ptr;
   USB_HOST_STATE_STRUCT_PTR usb_host_ptr;
   USB_STATUS status;



/*****************************************************************
CODE FOR ARC FULL SPEED CORE
*****************************************************************/

#ifdef __VUSB32__
#ifdef _HOST_DEBUG_
   DEBUG_LOG_TRACE("_usb_host_recv_data");
#endif

   usb_host_ptr = (USB_HOST_STATE_STRUCT_PTR)handle;

   pipe_descr_ptr = (PIPE_DESCRIPTOR_STRUCT_PTR)pipe_handle;
   
   /* Check if the pipe id is valid */
   if (pipe_descr_ptr->PIPE_ID > USB_MAX_PIPES) {
      #ifdef _HOST_DEBUG_
         DEBUG_LOG_TRACE("_usb_host_recv_data invalid pipe handle");
      #endif

      return USB_log_error(__FILE__,__LINE__,USBERR_INVALID_PIPE_HANDLE);
   } /* Endif */

   USB_lock();

   /* Check if a previously queued packet is still pending */
   if (pipe_descr_ptr->STATUS != USB_STATUS_IDLE) {
      USB_unlock();
      #ifdef _HOST_DEBUG_
         DEBUG_LOG_TRACE("_usb_host_recv_data transfer in progress");
      #endif
      return USB_STATUS_TRANSFER_IN_PROGRESS;
   } else {
      USB_unlock();
      /* Initialize the pipe desciptor with the new receive transfer request 
      ** for this pipe
      */
      pipe_descr_ptr->TR = *tr_params_ptr;
      pipe_descr_ptr->SOFAR = 0;
      pipe_descr_ptr->RX_PTR = tr_params_ptr->RX_BUFFER;

      if ((pipe_descr_ptr->PIPETYPE == USB_ISOCHRONOUS_PIPE) && 
         (pipe_descr_ptr->TODO1 > pipe_descr_ptr->MAX_PACKET_SIZE)) 
      {
         /* For isochronous pipe, we are accepting a maximum of MAX_PACKET_SIZE 
         ** transfer 
         */
         pipe_descr_ptr->TODO1 = pipe_descr_ptr->MAX_PACKET_SIZE;
      } else {
         pipe_descr_ptr->TODO1 = tr_params_ptr->RX_LENGTH;
      } /* Endif */

      #ifdef _DATA_CACHE_
      /*Must Flush and Invalidate the buffer before sending
      /receiving the data in it */
      USB_dcache_invalidate_mlines((void *)pipe_descr_ptr->RX_PTR,
                                    tr_params_ptr->RX_LENGTH);  
    
      #endif
      
      
      USB_lock();
      
      /* successful transfer initiation status. */
      pipe_descr_ptr->STATUS = USB_STATUS_TRANSFER_QUEUED;

#ifdef __USB_OS_MQX__

      /* Call the low-level routine to receive data */
      ((USB_CALLBACK_FUNCTIONS_STRUCT_PTR)usb_host_ptr->CALLBACK_STRUCT_PTR)->\
         HOST_RECV(handle, pipe_descr_ptr);
#else
      _usb_hci_vusb11_recv_data(handle, pipe_descr_ptr);
#endif
         
      pipe_descr_ptr->PACKETPENDING = TRUE;
      
      /* full speed driver assumes that transfer queue is succesfull. This should change
      in future releases */
      
      status = USB_STATUS_TRANSFER_QUEUED;

      USB_unlock();
   } /* Endif */


/*****************************************************************
CODE FOR ARC HS SPEED CORE
*****************************************************************/

#else
   PIPE_TR_STRUCT_PTR   pipe_tr_ptr;
   USB_STATUS  return_code;   

#ifdef _HOST_DEBUG_
   DEBUG_LOG_TRACE("_usb_host_recv_data");
#endif

   USB_lock();

   return_code = _usb_host_set_up_TR(pipe_handle, tr_params_ptr, &pipe_tr_ptr);

   if (return_code != USB_STATUS_TRANSFER_QUEUED) {
      USB_unlock();
      #ifdef _HOST_DEBUG_
         DEBUG_LOG_TRACE("_usb_host_recv_data transfer queue failed");
      #endif
      return USB_log_error(__FILE__,__LINE__,return_code);
   } /* Endif */
   
   usb_host_ptr = (USB_HOST_STATE_STRUCT_PTR)handle;
   pipe_descr_ptr = (PIPE_DESCRIPTOR_STRUCT_PTR)pipe_handle;

   #ifdef _DATA_CACHE_
   /*Must Flush and Invalidate the buffer before sending
   /receiving the data in it */
   USB_dcache_invalidate_mlines((void *)pipe_tr_ptr->RX_BUFFER,
                                 pipe_tr_ptr->RX_LENGTH);  
 
   #endif

   /* We have obtained the current TR on the Pipe's TR list 
   ** from _usb_host_set_up_TR
   */
#ifdef __USB_OS_MQX__

   /* Call the low-level routine to receive data */
   status = _usb_host_recv_data_call_interface (handle, pipe_descr_ptr, pipe_tr_ptr);
#else
   status = _usb_hci_vusb20_recv_data(handle, pipe_descr_ptr, pipe_tr_ptr);
#endif
         
   USB_unlock();
   
#endif
   #ifdef _HOST_DEBUG_
      DEBUG_LOG_TRACE("_usb_host_recv_data SUCCESSFUL");
   #endif

   if(status == USB_OK) 
   {
      return USB_STATUS_TRANSFER_QUEUED;
   }
   else
   {
      return USB_log_error(__FILE__,__LINE__,status);
   }

   
} /* Endbody */

/* EOF */


   

